

var URL = require("url")
, net = require("net")
, debug = require("../../lib/debug.js")("app:net")
, querystring = require("querystring")

, adapter = {
	http:  http,     // port 80
	https: http,     // port 443
	tcp: function(opts, next) {
		var socket = new net.Socket()
		, body = ""
		socket.connect(opts.port, opts.hostname, function() {
			if (opts.hParam.noDelay === "true") {
				socket.setNoDelay(true)
			}
			if (opts.body) {
				socket.write(opts.hParam.encoding ?
					Buffer.from(opts.body, opts.hParam.encoding) :
					opts.body
				)
			}
			socket.end()
		})
		socket.on("error", debug)
		socket.on("data", function(d) {
			body += d
		})
		socket.on("close", function() {
			if (next) {
				next(null, socket, body)
			}
		})
	},
	udp: function(opts, next) {
		var socket = require("dgram")
		.createSocket({
			reuseAddr: true,
			type: net.isIPv6(opts.hostname) ? "udp6" : "udp4"
		})
		socket.on("error", debug)

		socket.bind(function() {
			var message = opts.body
			, len = message.length
			if (opts.hParam.broadcast == "true") {
				socket.setBroadcast(true)
			}
			if (opts.hParam.ttl) {
				socket.setTTL(+opts.hParam.ttl)
			}
			if (opts.hParam.encoding) {
				message = Buffer.from(message, opts.hParam.encoding)
			}
			socket.send(message, opts.port, opts.hostname, function(err) {
				socket.close()
				if (next) {
					next(err, socket)
				}
			})
		})
	}
}


exports.request = request


function request(url, opts, next) {
	try {
		var req = validate(url)

		if (!next && typeof opts === "function") {
			next = opts
			opts = null
		}

		if (opts) {
			req.method = opts.method
			req.headers = opts.headers
			req.body = opts.body
		}

		debug(url)
		adapter[req.schema](req, next)
	} catch(e) {
		debug(e)
	}
}

function validate(url) {
	var req = typeof url === "string" ? URL.parse(url) : url
	, schema = req.schema = req.protocol.slice(0, -1)
	req.hParam = req.hash ? querystring.parse(req.hash.slice(1)) : {}
	if (!adapter[schema]) {
		throw "Invalid schema: " + req.schema
	}
	return req
}

function http(opts, next) {
	var req = require(opts.schema).request(opts, function(res) {
		var body = ""
		res.on("data", function(d) {
			body += d
		})
		res.on("end", function() {
			if (next) {
				next(null, res, body)
			}
		})
	})

	req.on("error", debug)

	if (opts.body) {
		req.write(opts.body)
	}

	req.end()
}

// PostgreSQL max query size 1 gigabyte
// SQLITE_MAX_SQL_LENGTH which defaults to 1000000
// mysql:
// SHOW VARIABLES LIKE 'max_allowed_packet';
// mysql max_allowed_packet=1048576 or 1mb


/*
Protocol	Purpose			Normal port	SSL variant	SSL port
SMTP		Send email		25/587		SMTPS		465 (legacy)[5][dead link]
POP3		Retrieve email		110		POP3S		995
IMAP		Read email		143		IMAPS		993
NNTP		News reader		119/433		NNTPS		563
LDAP		Directory Access	389		LDAPS		636
FTP		File transfer		21		FTPS		990
	// ftp:   require("./ftp"),    // port 21
	// imap://<iserver>/<enc-mailbox>[<uidvalidity>]<iuid>[<isection>][<ipartial>][<iurlauth>]
	// imap://michael@example.org/INBOX
	// imap://joe@example.com/INBOX/;uid=20/;section=1.2;urlauth=submit+fred:internal:91354a473744909de610943775f92038
	//imap:  require("./imap"),   // port 143
	// postgresql://host1:123,host2:456/somedb?target_session_attrs=any&application_name=myapp
	// smtps://address%40gmail.com:password@smtp.gmail.com
	//smtp:  require("./smtp"),   // port 25
	//ssdp:  require("./ssdp")    // port 1900
*/
