


!function(exports) {
	"use strict";
	var arrProto = Array.prototype
	, getFns = Object.create(null)
	, setFns = Object.create(null)
	, filterFns = Object.create(null)
	, escRe = /['\n\r\u2028\u2029]|\\(?!x2e)/g
	, intRe = /^-?\d+$/
	, pathRe = /(^$|.+?)(?:\[([^\]]*?)\])?(\.(?=[^.])|$)/g
	, reEscRe = /[.+^=:${}()|\/\\]/g
	, globRe = /[?*]|\[\w+\]/
	, globReplace = /\?|(?=\*)/g
	, globGroup = /\[!(?=.*\])/g
	, numRe = /^-?(\d*\.)?\d+$/
	, valRe = /(["'])(\\?.)*?\1|(\w*)\{((["'])(?:\\?.)*?\5|\w*\{(?:(["'])(?:\\?.)*?\6|[^}])*?\}|.)*?\}|(@?)[^,]+/g
	, filterRe = /(!?)(\$?)((?:[-+:.\/\w]+|\[[^\]]+?\])+)(\[])?(?:((?:!(?=\1)=|(?=\1)[<>=])=?)((?:(["'])(?:\\?.)*?\7|\w*\{(?:(["'])(?:\\?.)*?\8|\w*\{(?:(["'])(?:\\?.)*?\9|[^}])*?\}|.)*?\}|[^|&)])+))?|((&|\|)\11*|([()])|.)/g
	, onlyFilterRe = RegExp("^(" + filterRe.source.slice(0, -10) + "))+$")
	, cleanRe = /(\(o=d\)&&(?!.*o=o).*)\(o=d\)&&/g
	, fns = {
		"==": "e",
		"===": "s",
		">": "g",
		">=": "ge",
		"<": "l",
		"<=": "le"
	}
	, equal =  Fn("a->b->a==b")
	, strictEqual = Fn("a->b->a===b")
	, greater = Fn("a->b->a<b")
	, greaterEqual = Fn("a->b->a<=b")
	, less = Fn("a->b->a>b")
	, lessEqual = Fn("a->b->a>=b")
	, reMatch = Fn("a->b->typeof b==='string'&&a.test(b)")
	, tmpDate = new Date()

	exports.Item = Item
	exports.List = List

	Item.filterFn = filterFn
	Item.copy = copy
	Item.extend = List.extend = extend
	Item.cache = {}
	List.cache = {}

	Item.get = function(obj, pointer) {
		return pathFn(pointer)(obj)
	}
	Item.get.str = pathStr

	Item.set = function(obj, pointer, value) {
		return pathFn(pointer, true)(obj, value)
	}

	function escFn(str) {
		return escape(str).replace(/%u/g, "\\u").replace(/%/g, "\\x")
	}

	function extend(a, b, c, d, e) {
		var fn = this
		, clone = Function("return " + fn.toString())()
		, _super = fn.prototype
		clone.prototype = Object.assign(
			Array.isArray(_super) ? [] : Object.create(null),
			_super, a, b, c, d, e, {_super: _super, constructor: clone}
		)
		clone.extend = extend
		clone.cache = clone.prototype.cache || fn.cache
		return clone
	}

	function Item(attrs, opts) {
		var item = this
		if (attrs instanceof Item) {
			return attrs
		}
		if (!(item instanceof Item)) {
			return Item.cache[attrs.id] || (Item.cache[attrs.id] = new Item(attrs, opts))
		}
		item.data = attrs
		item.lists = []
		if (item.init) {
			item.init(attrs, opts)
		}
		item.set(attrs)
	}

	Item.prototype = {
		constructor: Item,
		set: function(key, val, opts) {
			var item = this
			, changed = []
			, previous = {}
			, data = key

			if (typeof key !== "object") {
				pathFn(key, true)(data = {}, val)
			} else {
				opts = val
			}

			if (item.validate(data, opts)) {
				JSON.mergePatch(item.data, data, changed, previous)
			}

			if (changed.length) {
				item.emit("change", changed, item, opts, previous)
				changed.each(function(pointer, idx) {
					item.emit("change:" + pointer, item.get(pointer), item, opts, previous[pointer])
				})
			}
			return changed
		},
		get: function(name, fallback) {
			var val = pathFn(name)(this.data)
			return val == void 0 ? fallback : val
		},
		validate: returnTrue,
		destroy: function(next) {
			var item = this
			, arr = item.lists.slice()
			, len = arr.length

			for (; len--; ) {
				arr[len].remove(item)
			}

			item.emit("destroy")
			item.off()
			item.constructor.cache[item.get("id")] = null
			//delete item.constructor.cache[item.get("id")]
		},
		matches: function(expr) {
			return filterFn(expr)(this.data)
		},
		then: _then,
		toJSON: function(attrs) {
			return Array.isArray(attrs) ? copy({}, this.data, attrs) : this.data
		}
	}
	Event.asEmitter(Item.prototype)

	function List(name, opts) {
		var list = this
		if (!(list instanceof List)) {
			return List.cache[name] || (List.cache[name] = new List(name, opts))
		}
		list.name = name
		if (list.init) {
			list.init(name, opts)
		}
		list.on("newListener", function(type, fn, scope, origin) {
			if (type.split(":")[0] === "change") {
				list.eachLive(add, remove)
				list.on("removeListener", off)
			}
			function add(item) {
				item.on(type, fn, scope, origin)
			}
			function remove(item) {
				item.off(type, fn, scope, origin)
			}
			function off(_type, _fn, _scope, _origin) {
				if (_type === type && _fn == fn && _scope == scope && _origin == origin) {
					list.each(remove).off("add", add).off("remove", remove)
					list.off("removeListener", off)
				}
			}
		})
	}

	Object.assign(List.prototype = [], Event.Emitter.prototype, {
		item: Item,
		wait: Fn.hold,
		syncMethods: [
			"add", "emit", "filterFn", "merge",
			"item", "paged", "pluck", "remove", "removeAll", "sortFn"
		],
		add: function(item) {
			var pos
			, list = this

			if (item) {
				if (item.constructor === Object) item = list.item(item)

				if (item.lists.indexOf(list) == -1) {
					pos = indexFor(list, item, list.sortFn)
					if (list.filterFn(item, pos)) {
						item.lists.push(list)
						list.splice(pos , 0, item)
						list.emit("add", item, pos)
					}
				}
			}
			return list
		},
		filterFn: returnTrue,
		remove: function(item) {
			var list = this
			if (item && item.lists && item.lists.remove(list) > -1) {
				list.emit("remove", item, arrProto.remove.call(list, item))
			}
			return list
		},
		removeAll: function(){
			for (var list = this, len = list.length; len--; ) {
				list.remove(list[len])
			}
			return list
		},
		each: function(fn, scope) {
			arrProto.each.call(this, fn, scope || this)
			return this
		},
		eachLive: function(start, stop, scope) {
			this
			.on("add", start, scope)
			.on("remove", stop, scope)
			.slice(0)
			.each(start, scope)
			return this
		},
		get: function(id, next, scope) {
			var list = this
			, item = list.item.cache[id]

			if (item && item.lists.indexOf(list) != -1) {
				return next ?
					(next.call(scope || list, null, item), list) :
					item
			}

			next && next.call(scope || list, Error("Item not found"))
		},
		at: function(index, next, scope) {
			var list = this
			, item = list[index < 0 ? list.length + index : index]
			, err = item ? null : Error("Item not found")

			return next ?
				(next.call(scope || list, err, item), list) :
				item
		},
		then: _then,
		pluck: function(field, next, scope) {
			var list = this
			, arr = arrProto.map.call(list, function(item) {
				return item.get(field)
			})
			return next ?
				(next.call(scope || list, arr), list) :
				arr
		},
		set: function(key, val) {
			this.each(function(item) {
				item.set(key, val)
			})
			return this
		},
		merge: function(mergeList) {
			var list = this
			, lists = list.lists || (list.lists = [])

			if (mergeList && lists.indexOf(mergeList) == -1) {
				lists.push(mergeList)

				mergeList
				.eachLive(list.add, _remove, list)
				.on("change", _change, list)
				.then(list.wait())
			}

			return list
		},
		unmerge: function(mergedList) {
			var list = this
			, lists = list.lists
			, idx = lists && lists.indexOf(mergedList)

			if (lists && mergedList && idx != -1) {
				lists.splice(idx, 1)

				mergedList
				.each(_remove, list)
				.off("add", list.add, list)
				.off("remove", _remove, list)
				.off("change", _change, list)
			}
			return list
		},
		paged: function(num) {
			var list = this
			, view = new List()
			, start = 0
			, end = num

			view.filterFn = function(item, pos) {
				pos = list.indexOf(item)
				return pos >= start && pos < end
			}

			view.pages = new List()
			view.setPage = function(newPage) {
				var lastPage = 0 | (list.length / num)
				if (newPage > lastPage) {
					newPage = lastPage
				}
				view.page = newPage
				start = num * newPage
				end = start + num
				view.reFilter()
			}

			view.merge(list)

			return view
		},
		reFilter: function() {
			var list = this
			, lists = list.lists || (list.lists = [])

			lists.each(function(parent) {
				parent.each(function(item) {
					_change.call(list, null, item)
				})
			})
		},
		reSort: function(fn) {
			var item
			, list = this
			, len = list.length
			, removed = []

			if (typeof fn === "string") {
				fn = [fn]
			}

			if (Array.isArray(fn)) {
				fn = Fn("a b->" + fn.map(function(key) {
					return "(A=a.get('" + key + "'))<(B=b.get('" + key + "'))?-1:A>B?1:0"
				}).join("||"))
			}

			fn = list.sortFn = fn || list.sortFn

			for (; len > 1; ) {
				item = list[--len - 1]
				if (fn(item, list[len]) > 0) {
					removed.push(item)
					list.remove(item)
				}
			}
			removed.each(list.add, list)
		},
		toJSON: function() {
			return this.slice(0)
		}
	})

	function _remove(item) {
		if (item.lists.every(notIn, this.lists)) {
			this.remove(item)
		}
	}

	function _change(changed, item) {
		var list = this
		, matches = list.filterFn(item)
		, inList = item.lists.indexOf(list) > -1

		if (matches && !inList) {
			list.add(item)
		}
		if (!matches && inList) {
			list.remove(item)
		}
	}


	function pathStr(str, set) {
		return (
			str.charAt(0) === "/" ?
			str.slice(1).replace(/\./g, "\\x2e").replace(/\//g, ".").replace(/~1/g, "/").replace(/~0/g, "~") :
			str
		)
		.replace(escRe, escFn)
		.replace(pathRe, set === true ? pathSet : pathGet)

	}
	function pathGet(str, path, arr, dot) {
		if (arr && !onlyFilterRe.test(arr)) throw Error("Invalid filter in GET")
		var v = dot ? "(o=" : "(c="
		return (
			arr ? pathGet(0, path, 0, 1) + "i(o)&&" + v + (
				intRe.test(arr) ? "o[" + (arr < 0 ? "o.length" + arr : arr) + "])" :
				"F(o,f('" + arr + "')))"
			) :
			v + "o['" + path + (arr === "" ? "'])&&i(c)&&c" : "'])")
		) + (dot ? "&&" : "")
	}
	function pathSet(str, path, _arr, dot) {
		var op = "o['" + path + "']"
		, arr = _arr || ""
		if (arr) {
			arr = "(o=i(o['" + path + "'])?o['" + path + "']:(o['" + path + "']=[]))&&"
			if (intRe.test(_arr)) {
				op = "o[" + (_arr < 0 ? "o.length" + _arr : _arr) + "]"
			} else if (_arr === "-") {
				op = "o[o.length]"
			} else {
				if (!onlyFilterRe.test(_arr)) throw Error("Invalid filter in SET")
				op = "o[t]"
				arr += "(t=F(o,f('" + _arr + "'),1))!=null&&"
			}
		}
		return (dot ?
			arr + "(o=typeof " + op + "==='object'&&" + op + "||(" + op + "={}))&&" :
			arr + "((c=" + op + "),(" + op + "=v),c)"
		)
	}
	function pathFn(str, set) {
		var map = set === true ? setFns : getFns
		return (
			map[str] ||
			(map[str] = Function("i,f,F", "return function(d,v){var c,o,t;return (o=d)&&" + pathStr(str, set) + "||c}")(
				Array.isArray, filterFn, findFn
			))
		)
	}


	function copy(a, b, attrs) {
		var len = b && attrs && attrs.length
		return len ? (
			copy[len] ||
			(copy[len] = Function("a,b,k,g", "var v,u;return " + Object.keys(attrs).map(copyFnI) + ",a"))
		)(a, b, attrs, pathFn) : a
	}

	function copyFnI(i) {
		return "(v=g(k[" + i + "])(b))!==u&&g(k[" + i + "],true)(a,v)"
	}

	function returnTrue() {
		return true
	}

	function _then(next, scope) {
		if (typeof next === "function") {
			next.call(scope || this)
		}
		return this
	}

	function notIn(v) {
		return this.indexOf(v) == -1
	}

	function indexFor(arr, el, compFn) {
		var o
		, i = 0
		, l = arr.length

		// Fast lookup for last position
		if (compFn && l && compFn(el, arr[l-1]) < 1) {
			// Otherwise binary search
			for (; i < l; ) {
				compFn(el, arr[o=(i+l)>>1]) < 0 ? l=o : i=o+1
			}
		}
		return l
	}

	function findFn(a, fn, idx) {
		for (var i = -1, len = a.length; ++i < len; ) {
			if (fn(a[i])) return idx == null ? a[i] : i
		}
		return idx != null && len
	}

	function filterFn(str, prefix, opts, getter, tmp) {
		var optimized
		, arr = []
		, key = (prefix || "") + filterStr(str, opts, arr, getter)
		, fn = filterFns[key]
		if (!fn) {
			for (optimized = key; optimized != (optimized = optimized.replace(cleanRe, "$1")); );
			fn = filterFns[key] = Function("a,i,f,p,e,s,g,ge,l,le,m,F,t", "return function(d){var o;return "+optimized+"}")
		}
		return fn(
			arr, Array.isArray, filterFn, pathFn,
			equal, strictEqual, greater, greaterEqual, less, lessEqual, reMatch, findFn, tmp
		)
	}
	filterFn.str = filterStr
	filterFn.valRe = valRe

	// Date{day=1,2}
	// Geo{distance=200km&lat=40&lon=-70}
	// ?pos=Geo{distance=200km&lat=@lat&lon=@lon}
	// [lon, lat] in The GeoJSON Format RFC 7946
	// IP{net=127.0.0.1/30}
	// var re = /^((\d|[1-9]\d|1\d\d|2[0-4]\d|25[0-5])(?:\.(?=.)|$)){4}$/
	// ["1.2.3.4", "127.0.0.1", "192.175.255.254."].map(re.test, re)

	filterFn.date = function filterDate(str) {
		return filterFn(str, "(t.setTime(+d)>=0)&&", null, dateGetter, tmpDate)
	}
	function dateGetter(name) {
		return "(t.get" + Date.fnMap[name] + ")"
	}

	function filterStr(qs, opts, arr, getter) {
		return qs.replace(filterRe, worker).replace(/^[1&]&+|&+1?$/g, "") || "1"

		function worker(all, not, isOption, attr, isArray, op, val, q1, q2, q3, ext, ok2, ok1) {
			if (ext) {
				if (!ok2 && !ok1) {
					throw new Error("Invalid filter: " + qs)
				}
				return ok2 ? ok2 + ok2 : ok1
			}
			if (isOption) {
				if (opts) opts[attr] = val
				return "1"
			}

			var idd, m, v, isRe
			, a = []
			, pre = "(o=d)&&"
			, not2 = op == "!=" || op == "!=="

			attr = (getter || pathStr)(attr).replace(/\(c=(.*?)\)$/, "$1")

			if (not2) {
				op = op.slice(1)
			}

			if (!op) {
				return pre + (not ? "!" : "") + (isArray ? "i(" + attr + ")" : attr)
			}
			if (op == "=" || op == "==") op += "="
			for (; m = valRe.exec(val); ) {
				isRe = 0
				idd = arr.indexOf(v =
					m[1] || m[4] ? m[0].slice(m[3] ? m[3].length + 1 : 1, -1) :
					m[7] ? m[0].slice(1) :
					numRe.test(m[0]) ? +m[0] :
					(isRe = globRe.test(m[0])) ? RegExp(
						"^" + m[0]
						.replace(reEscRe, "\\$&")
						.replace(globReplace, ".")
						.replace(globGroup, "[^") + "$",
						op === "==" ? "i" : ""
					) :
					m[0]
				)
				v = "a[" + (idd > -1 ? idd : arr.push(v) - 1) + "]"
				idd = m[3] ? 'f.' + m[3].toLowerCase() : m[4] ? "f" : isRe ? "m" : fns[op]
				a.push(
					isArray ? attr + ".some(" + idd + "(" + v + "))" :
					isRe ? "typeof " + attr + "==='string'&&" + v + ".test(" + attr + ")" :
					m[3] || m[4] ? idd + '(' + v + ")(" + attr + ")" :
					m[7] ? attr + "!==void 0&&" + attr + op + "p(" + v + ")(o)" :
					attr + op + v
				)
			}

			return pre + (isArray ? (not2 ? "!i("+attr+")||" : "i("+attr+")&&") : "") + (not2 ? "!(" : "(") + a.join("||") + ")"
		}
	}
}(this)



