
!function(window, xhr) {
	var imageType = /image.*/
	, iframeCount = 0
	, canvas = document.createElement("canvas")
	, ctx = canvas.getContext && canvas.getContext("2d")

	// Detect the Android stock browser via the UA string
	// as Blob sent with XMLHttpRequest2 there results an empty body.
	// The work-around would be to send an ArrayBuffer: https://issuetracker.google.com/issues/36958619
	// NOTE:2015: most Android devices now ship with Chrome by default
	, stockVer = +/android.+webkit\/([0-9]+)(?!.*chrome)|$/i.exec(navigator.userAgent)[1]
	, shouldWork = (!stockVer || stockVer > 536) && ctx && window.Blob && window.FormData && window.FileReader

	function dataURLtoBlob(dataURL) {
		var comma = dataURL.indexOf(",")
		, head = dataURL.slice(5, comma)
		, mime = head.split(";")[0]
		, byteString = (head.slice(-6) == "base64" ? atob : decodeURI)(dataURL.slice(comma + 1))
		, len = byteString.length
		, ab = new ArrayBuffer(len)
		, ia = new Uint8Array(ab)
		for (; len--;) {
			ia[len] = byteString.charCodeAt(len)
		}
		try {
			return new Blob([ab], { type: mime })
		} catch(e) {
			var Builder = window.BlobBuilder ||
				window.WebKitBlobBuilder ||
				window.MozBlobBuilder ||
				window.MSBlobBuilder
			, blob = new Builder()
			blob.append([ab.buffer])
			return blob.getBlob(mime)
		}
	}

	function resize(img, maxWidth, maxHeight, crop) {
		var type = img.type || "image/jpeg"
		, quality = quality || 0.92
		, ratio = maxWidth / maxHeight
		, imgRatio = img.naturalWidth / img.naturalHeight

		canvas.width = maxWidth
		canvas.height = maxWidth / imgRatio

		// TODO:2014-06-27:lauri:Use down-stepping to achive better results.
		ctx.drawImage(img, 0, 0, img.naturalWidth, img.naturalHeight, 0, 0, canvas.width, canvas.height)

		return canvas.toDataURL(type, quality)
	}

	function upload(form, callback) {
		var key, el, file, formData, fileReader
		, pending = 0
		, i = 0

		// Do form submit as fallback
		if (!shouldWork) return false

		formData = new FormData()

		function appendFile(key, dataURL, name) {
			var match = dataURL.match(/^data:\w+?\/(\w+)/)
			, ext = match && match[1] || "jpg"
			formData.append(key, dataURLtoBlob(dataURL), name.replace(/(\.\w+)?$/, "." + ext) )

			if (--pending == 0) {
				xhr("POST", form.action, callback).send(formData)
			}
		}

		for (; el = form.elements[i++];) if ((key = el.name || el.id) && !el.disabled) {
			if (el.type != "file") {
				// TODO:2014-06-27:lauri:Add radio and checkbox support
				formData.append(key, el.value)
				continue
			}
			// TODO:2014-10-21:lauri:loop all files
			file = el.files && el.files[0]
			if (!file) return false

			fileReader = new FileReader()

			pending++

			fileReader.onload = (function(el, key) {
				var img = document.createElement("img")
				, maxWidth = el.getAttribute("maxwidth")

				img.onload = function() {
					var dataURL = img.naturalWidth > maxWidth && resize(img, maxWidth) || img.src
					appendFile(key, dataURL, file.name || "file")
				}

				return function(e) {
					var dataURL = e.target.result
					if (maxWidth && file.type.match(imageType)) {
						img.src = e.target.result
					} else {
						appendFile(key, dataURL)
					}
				}
			})(el, key, file)

			fileReader.readAsDataURL(file);
		}
		return true
	}

	function uploadResized(form, callback) {
		var uploaded = upload(form, callback)
		if (!uploaded) {
			var id = form.target = "lt_upl_ifr" + (iframeCount++)
			, iframe = El("iframe.hidden[name=" + id + "]")
			El.on(iframe, "load", iframeLoad)
			El.to(iframe, document.body)
		}
		return !uploaded

		function iframeLoad() {
			var iframe = this
			El.off(iframe, "load", iframeLoad)
			// .getElementsByTagName('body')[0];
			var iframeBody = (iframe.contentDocument || iframe.contentWindow.document || iframe.document).body
			, text = iframeBody.textContent || iframeBody.innerText
			callback(!text, text)
			El.kill(iframe)
		}
	}

	uploadResized.submit = function(el, callback) {
		// input.form Chrome9 Firefox4 IE10 Opera10.62
		var form = el.form || El.closest(el, "form")
		return uploadResized(el, callback) && el.submit()
	}
	xhr.uploadResized = uploadResized
}(this, xhr)


